// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2023 Oracle.  All Rights Reserved.
 * Author: Darrick J. Wong <djwong@kernel.org>
 */
#ifndef __LIBFROG_RANDBYTES_H__
#define __LIBFROG_RANDBYTES_H__

extern uint8_t randbytes_test_buf[];

#endif /* __LIBFROG_RANDBYTES_H__ */
