// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2006 Silicon Graphics, Inc.
 * All Rights Reserved.
 */
#ifndef LIBXFS_INIT_H
#define LIBXFS_INIT_H

struct stat;
extern int     use_xfs_buf_lock;

#endif	/* LIBXFS_INIT_H */
